import { IServiceProvider } from "./IServiceProvider.js";
import { ServiceCollection } from "./ServiceCollection.js";
import { ServiceRegistry } from "./ServiceRegistry.js";
import { getNameOfServiceType } from "../helpers/getNameOfServiceType.js";
import { ServiceInstancesStore } from "./ServiceInstancesStore.js";
import { ServiceLifetime } from "./ServiceLifetime.js";
import { isAsyncServiceImplementationLoader } from "../helpers/isAsyncServiceImplementationLoader.js";
import { validateRequestPath } from "./validation/validateRequestPath.js";
import { getNameOfDescriptor } from "../helpers/getNameOfDescriptor.js";
import { Debug } from "../debug.js";
export class ServiceProvider {
    collection;
    parent;
    static internalGetService(provider, service, requestedBy) {
        if (!(provider instanceof ServiceProvider)) {
            throw new Error("provider must be an instance of ServiceProvider");
        }
        return provider.internalGetService(service, requestedBy);
    }
    static getDescriptor(provider, service) {
        if (!(provider instanceof ServiceProvider)) {
            throw new Error("provider must be an instance of ServiceProvider");
        }
        return provider.getDescriptor(service);
    }
    static getDescriptors(provider, service) {
        if (!(provider instanceof ServiceProvider)) {
            throw new Error("provider must be an instance of ServiceProvider");
        }
        return provider.collection.getDescriptors(service);
    }
    instancesStore;
    constructor(collection, parent) {
        this.collection = collection;
        this.parent = parent;
        this.instancesStore = new ServiceInstancesStore();
    }
    getServices(service) {
        return this.resolveGeneratorSync(this.internalGetService([service], new Set()));
    }
    getServiceAsync(service) {
        return this.resolveGeneratorAsync(this.internalGetService(service, new Set()));
    }
    getService(service) {
        return this.resolveGeneratorSync(this.internalGetService(service, new Set()));
    }
    getServicesAsync(service) {
        return this.resolveGeneratorAsync(this.internalGetService([service], new Set()));
    }
    createScope() {
        const serviceProvider = new ServiceProvider(this.collection, this);
        return {
            serviceProvider,
            [Symbol.dispose]: () => {
                serviceProvider[Symbol.dispose]();
            },
        };
    }
    createAsyncScope() {
        const serviceProvider = new ServiceProvider(this.collection, this);
        return {
            serviceProvider,
            [Symbol.asyncDispose]: async () => {
                await serviceProvider[Symbol.asyncDispose]();
            },
        };
    }
    *internalGetService(service, requestedBy) {
        const isAllServices = Array.isArray(service);
        const descriptors = isAllServices
            ? this.collection.getDescriptors(service[0])
            : [this.getDescriptor(service)];
        const services = [];
        for (const descriptor of descriptors) {
            services.push(yield* this.internalGetDescriptorImplementation(descriptor, requestedBy));
        }
        if (isAllServices) {
            return services;
        }
        return services[0];
    }
    *internalGetDescriptorImplementation(descriptor, requestedBy) {
        if (descriptor.lifetime === ServiceLifetime.Singleton &&
            this.parent &&
            descriptor.service !== IServiceProvider) {
            return yield* this.parent.internalGetDescriptorImplementation(descriptor, requestedBy);
        }
        validateRequestPath(requestedBy, descriptor);
        return yield* this.createInstanceFromDescriptor(descriptor, requestedBy);
    }
    *createInstanceFromDescriptor(descriptor, requestedBy) {
        try {
            const lastRequestedBy = [...requestedBy].pop();
            if (lastRequestedBy?.lifetime !== undefined) {
                if (lastRequestedBy.lifetime === ServiceLifetime.Singleton &&
                    descriptor.lifetime === ServiceLifetime.Scoped) {
                    throw new Error(`Scoped service cannot be resolved from singleton service.`);
                }
            }
            switch (descriptor.lifetime) {
                // @ts-expect-error
                case ServiceLifetime.Scoped:
                    if (!this.parent) {
                        throw new Error("Scoped services require a service scope.");
                    }
                case ServiceLifetime.Singleton: {
                    const instanceInfo = this.instancesStore.addInstance(descriptor, lastRequestedBy);
                    instanceInfo.initialize(yield* this.createServiceInitializer(descriptor, requestedBy));
                    return instanceInfo.instance;
                }
                case ServiceLifetime.Transient:
                    return (yield* this.createServiceInitializer(descriptor, requestedBy))();
            }
        }
        catch (exception) {
            throw new Error(`Failed to initiate service ${getNameOfDescriptor(descriptor)}:\n\r${exception.message}`, { cause: exception });
        }
    }
    *createServiceInitializer(descriptor, requestedBy) {
        let implementation = descriptor.implementation;
        if (isAsyncServiceImplementationLoader(implementation)) {
            if (!implementation.isLoaded()) {
                yield implementation.load();
            }
            implementation = implementation.getImplementation();
        }
        const metadata = ServiceRegistry.get(implementation);
        if (metadata) {
            const dependencies = [];
            requestedBy = new Set([...requestedBy, descriptor]);
            for (const dependency of metadata.dependencies) {
                dependencies.push(yield* this.internalGetService(dependency, requestedBy));
            }
            return () => new implementation(...dependencies);
        }
        if (descriptor.service === IServiceProvider) {
            return () => this;
        }
        if (typeof implementation === "function") {
            return () => {
                try {
                    return implementation(this);
                }
                catch (err) {
                    if (err instanceof TypeError &&
                        err.message.includes("cannot be invoked without 'new'")) {
                        throw new Error(Debug.errors.classNotDecorated(implementation.name), {
                            cause: err,
                        });
                    }
                    else {
                        throw err;
                    }
                }
            };
        }
        return () => implementation;
    }
    getDescriptor(service) {
        const descriptors = this.collection.getDescriptors(service);
        if (descriptors.length === 0) {
            let name = getNameOfServiceType(service);
            const metadata = ServiceRegistry.get(service);
            if (metadata?.name) {
                name = metadata.name;
            }
            throw new Error(`No service of type "${name}" is registered`);
        }
        return descriptors[descriptors.length - 1];
    }
    resolveGeneratorSync(iterator) {
        let result;
        while (!(result = iterator.next()).done) {
            throw new Error(`Lazy service cannot be resolved synchronously`);
        }
        return result.value;
    }
    async resolveGeneratorAsync(iterator) {
        let result;
        while (!(result = iterator.next()).done) {
            await result.value;
        }
        return result.value;
    }
    [Symbol.dispose]() {
        this.instancesStore[Symbol.dispose]();
    }
    async [Symbol.asyncDispose]() {
        await this.instancesStore[Symbol.asyncDispose]();
    }
}
//# sourceMappingURL=ServiceProvider.js.map